# Copyright (c) 2009 gocept gmbh & co. kg
# See also LICENSE.txt
"""Modelling network artefacts."""

import gocept.directory.consistency
import gocept.directory.form
import gocept.directory.interfaces
import gocept.reference
import grok
import megrok.pagelet
import netaddr
import zc.sourcefactory.basic
import zope.component
import zope.interface


class Network(grok.Container):

    grok.implements(gocept.directory.interfaces.INetwork)

    _cidr = None

    def __init__(self, cidr=None):
        super(Network, self).__init__()
        self.cidr = cidr

    def set_cidr(self, value):
        if self._cidr is not None:
            raise RuntimeError('Changing CIDR is not supported.')
        self._cidr = value
    cidr = property(fset=set_cidr, fget=lambda self:self._cidr)

    @property
    def twins(self):
        directory = self.vlan.location.directory
        for network in directory.networks:
            if network is self:
                continue
            if self.netaddr in network.netaddr:
                yield network
            elif network.netaddr in self.netaddr:
                yield network

    @property
    def netaddr(self):
        return netaddr.CIDR(self.cidr)

    @property
    def vlan(self):
        return self.__parent__

    @property
    def addresses(self):
        addresses = []
        for obj in self.values():
            if gocept.directory.interfaces.IAddress.providedBy(obj):
                addresses.append(obj)
        return sorted(addresses)

    @property
    def addresses_in_use(self):
        for addr in self.addresses:
            if addr.resource is None:
                continue
            yield addr

    def allocate(self, resource, address=None):
        """Allocate an address for the given resource."""
        if address is None:
            index = 2
            address = Address()
            address.ip = str(self.netaddr[index])
            while self.in_use(address):
                index += 1
                address.ip = str(self.netaddr[index])

        if self.in_use(address):
            raise RuntimeError('Address already allocated: %s in %s vlan %s' %
                               (address.ip, self.cidr, self.vlan.class_.short))

        if not address.netaddr in self.netaddr:
            raise RuntimeError('Address not in network.')

        address.resource = resource
        if address.network is not None and address.network != self:
            del address.network[address.__name__]
        self[address.ip] = address

        return address

    def in_use(self, address):
        """Determine whether an address is being used."""
        # We need to check for all addresses with the same IP, thus a simple
        # test for this address object isn't sufficient. We also have to go
        # through the others and check whether they're in use.
        if address.resource is not None:
            return True

        nets = [self] + list(self.twins)
        for net in nets:
            if address.ip in net and net[address.ip].resource is not None:
                return True

        return False

    def unused(self):
        used = 0
        for addr in self.addresses:
            if addr.resource is not None:
                used += 1
        return self.netaddr.size() - used


class AddNetwork(gocept.directory.form.BasicAddForm):

    factory = Network

    grok.context(gocept.directory.interfaces.IVLAN)
    form_fields = grok.AutoFields(gocept.directory.interfaces.INetwork)

    def chooseName(self, obj):
        return obj.cidr.replace('/', '-')


class Index(megrok.pagelet.Pagelet):

    grok.context(gocept.directory.interfaces.INetwork)
    grok.require('gocept.directory.ManageDirectory')


class Address(grok.Model):

    grok.implements(gocept.directory.interfaces.IAddress)

    resource = gocept.reference.Reference()

    def __init__(self, ip=None):
        super(Address, self).__init__()
        self.resource = None
        self.ip = ip

    @property
    def cidr(self):
        return '%s/%s' % (self.ip, self.network.netaddr.prefixlen)

    @property
    def network(self):
        return self.__parent__

    @property
    def netaddr(self):
        return netaddr.IP(self.ip)

    def __cmp__(self, other):
        if isinstance(other, Address):
            return cmp(self.netaddr, other.netaddr)
        return 0


class AddressConsistency(gocept.directory.consistency.UniqueResource):

    type_ = 'IP'

    def resources(self):
        for network in self.directory.networks:
            for address in network.addresses_in_use:
                yield address.ip, address.resource.title


class Deallocate(grok.View):

    grok.context(Address)

    def update(self):
        address = self.context
        resource = address.resource
        if isinstance(self.context.resource, gocept.directory.nodes.Node):
            # This is legacy code where an import went wrong
            for nic in resource.nics:
                if address in nic.addresses:
                    resource = nic
                    break
            else:
                raise RuntimeError('Address not found on associated node')
        resource.addresses.remove(self.context)
        address.resource = None

    def render(self):
        self.redirect(self.url(self.context))


class AddressIndex(megrok.pagelet.Pagelet):

    grok.name('index')
    grok.context(Address)
